# space_view3d_gta_tools
# 2011 ponz
script_info = "GTA Tools __init.py__ ( built:2012.02.05 )"

# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or ( at your option ) any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, 
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

#####
# ToDo.
#  - 
#####

bl_info = {
	"name": "GTA Tools", 
	"author": "ponz", 
	"version": ( 0, 20111211 ), 
	"blender": ( 2, 6, 0 ), 
	"api": 35853, 
	"location": "View3D > Tool Shelf > GTA Tools Panel", 
	"description": "Tools for GTA SA Models/Animations", 
	"warning": "", 
	"wiki_url": "", 
	"tracker_url": "", 
	"category": "3D View"}


if "bpy" in locals():
	import imp
	imp.reload(import_dff)
	imp.reload(export_dff)
	imp.reload(import_ifp)
	imp.reload(export_ifp)
	imp.reload(import_ipl)
	imp.reload(extract_txd)
	imp.reload(weight_tools)
	imp.reload(gta_utils)
else:
	from . import import_dff, export_dff, import_ifp, export_ifp, import_ipl, extract_txd, weight_tools, gta_utils

import bpy
import os
import time
import struct
from bpy.props import *


## Classes

### DFF_Import / DFF_Export ###
# Data Class

class GTA_DFF_Props(bpy.types.PropertyGroup):
	@classmethod
	def register( GTA_DFF_Props ):
		GTA_DFF_Props.imp_filepath = StringProperty(
			name="DFF File Path",
			description="Full Path Name of Import DFF",
			default = "")
		
		GTA_DFF_Props.imp_type = EnumProperty(
			name="Type",
			items = (
				("OTHER",   "Other", ""),
				("VEHICLE", "Vehicle", ""),
				("CHAR",    "Character", "") ),
			description="",
			default = "CHAR")
		
		GTA_DFF_Props.char_skin = BoolProperty(
			name="as char skin",
			description="As CHAR Skin",
			default = True)
		
		GTA_DFF_Props.vehicle = BoolProperty(
			name="as vehicle",
			description="As Vehicle",
			default = True)
		
		GTA_DFF_Props.map = BoolProperty(
			name="as map",
			description="As map",
			default = True)
		
		GTA_DFF_Props.ren_bone = BoolProperty(
			name="Rename Bone",
			description="Rename bone for Mirroring Modifire in Blender",
			default = True)
		
		GTA_DFF_Props.use_msplit = BoolProperty(
			name="Use Material Split",
			description="Assign Sub-Material ID to each Faces.",
			default = True)
		
		GTA_DFF_Props.use_remdbls = BoolProperty(
			name="Weld By Normal",
			description="Weld vertices doubled(very close) in both coordinate space and normal space.",
			default = True)
		
		GTA_DFF_Props.remdbls_th_co = FloatProperty(
			name="co",
			description="Matching length used in Remove Doubled for coordinate space.",
			min = .0001, max = 2.0, default = 0.0001, step = 0.01,
			precision = 4 )
		
		GTA_DFF_Props.remdbls_th_no = FloatProperty(
			name="no",
			description="Matching length used in Remove Doubled for normal space.",
			min = .0001, max = 2.0, default = 0.0001, step = 0.01,
			precision = 4 )
		
		GTA_DFF_Props.exp_filepath = StringProperty(
			name="Exoprt DFF File Path",
			description="Full Path Name of Export DFF",
			default = "")
		
		GTA_DFF_Props.exp_type = EnumProperty(
			name="Type",
			items = (
				("OTHER",   "Other", ""),
				("VEHICLE", "Vehicle", ""),
				("CHAR",    "Character", "") ),
			description="",
			default = "CHAR")
		
		GTA_DFF_Props.uv_th = FloatProperty(
			name="UV threshold",
			description="Matching Threshold for UV coods",
			min = 0.0, default = 0.0001,
			precision = 4 )
		
		GTA_DFF_Props.rev_bone = BoolProperty(
			name="Revert Bone Name",
			description="Revert bone name to original DFF bone",
			default = True)
		
		GTA_DFF_Props.dump_stat = BoolProperty(
			name="dump stat",
			description="dump dff import status",
			default = False)
		
		GTA_DFF_Props.use_alphatex = BoolProperty(
			name="Use Alpha Texture",
			description="Use Alpha Textures named [texture name]+\"a\".",
			default = True)
		
		GTA_DFF_Props.tex_path = StringProperty(
			name="Texture Path",
			description="",
			default = "")

bpy.utils.register_class( GTA_DFF_Props )


# Operators

class OperatorImportDff( bpy.types.Operator ):
	bl_idname = "import_dff.imp_dff"
	bl_label = "Select DFF file (*.dff)"
	
	filepath = bpy.props.StringProperty( subtype = "FILE_PATH" )
	filename_ext = ".dff"
	filter_glob = StringProperty( default = "*.dff", options = {'HIDDEN'} )
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.set_time_ini()
		gta_tools.msg = "--- DFF Import ---\n"
		gta_tools.dff_props.imp_filepath = self.filepath
		gta_tools.dff_props.map = False
		
		from . import import_dff
		import_dff.ImportDFF()
		gta_tools.show_msg_elapsed()
		
		return {'FINISHED'}
	
	def invoke( self, context, event ):
		gta_tools = bpy.context.scene.gta_tools
		self.filepath = gta_tools.dff_props.imp_filepath
		context.window_manager.fileselect_add( self )
		return {'RUNNING_MODAL'}

class OperatorExportDff( bpy.types.Operator ):
	bl_idname = "export_dff.exp_dff"
	bl_label = "Select DFF file (*.dff)"
	
	filepath = bpy.props.StringProperty( subtype = "FILE_PATH" )
	filename_ext = ".dff"
	filter_glob = StringProperty( default = "*.dff", options = {'HIDDEN'} )
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.set_time_ini()
		gta_tools.msg = "--- DFF Export ---\n"
		gta_tools.dff_props.exp_filepath = self.filepath
		
		from . import export_dff
		export_dff.ExportDFF( self.filepath )
		gta_tools.show_msg_elapsed()
		return {'FINISHED'}
	
	def invoke( self, context, event ):
		gta_tools = bpy.context.scene.gta_tools
		self.filepath = gta_tools.dff_props.exp_filepath
		context.window_manager.fileselect_add( self )
		return {'RUNNING_MODAL'}



# UI Panel

class GTA_DFFIO_UI( bpy.types.Panel ):
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'TOOLS'
	bl_context = "objectmode"
	bl_label = "GTA DFF Tools"
	
	def draw( self, context ):
		layout = self.layout
		
		if bpy.context.mode == 'OBJECT':
			gta_tools = bpy.context.scene.gta_tools
			col = layout.column()
			col.row().label( text = "Import DFF:" )
			col.row().operator( "import_dff.imp_dff", "Import" )
			col.row().prop( gta_tools.dff_props, "imp_type" )
			if "CHAR" == gta_tools.dff_props.imp_type:
				col.row().prop( gta_tools.dff_props, "ren_bone" )
			#if "VEHICLE" == gta_tools.dff_props.imp_type:
			#	col.row().prop( gta_tools.dff_props, "imp_vehicle_coll" )
			
			col.row().prop( gta_tools.dff_props, "use_msplit" )
			col.row().prop( gta_tools.dff_props, "use_remdbls" )
			if gta_tools.dff_props.use_remdbls:
				col = layout.column( align = True )
				col.row().prop( gta_tools.dff_props, "remdbls_th_co" )
				col.row().prop( gta_tools.dff_props, "remdbls_th_no" )
			#col.row().prop( gta_tools.dff_props, "use_map_normal" )
			#if gta_tools.dff_props.use_map_normal:
			#	col.row().prop( gta_tools.dff_props, "normal_factor" )
			
			col = layout.column()
			col.row().label( text = "Export DFF:" )
			col.row().operator( "export_dff.exp_dff", "Export" )
			col.row().prop( gta_tools.dff_props, "exp_type" )
			col.row().prop( gta_tools.dff_props, "uv_th" )
			if "CHAR" == gta_tools.dff_props.exp_type:
				col.row().prop( gta_tools.dff_props, "rev_bone" )





### MAP Import ###
# Data Class

class GTA_IPL_Props(bpy.types.PropertyGroup):
	@classmethod
	def register( GTA_IPL_Props ):
		from bpy.props import PointerProperty, StringProperty, BoolProperty, EnumProperty, IntProperty, CollectionProperty
		
		GTA_IPL_Props.name = StringProperty(
			name="IPL Name",
			description="",
			default = "")
		
		GTA_IPL_Props.ipl_name = StringProperty(
			name="IPL Name, Num Insts, Num Bin Insts",
			description="",
			default = "")
		
		GTA_IPL_Props.path = StringProperty(
			name="IPL filepath",
			description="",
			default = "")

bpy.utils.register_class( GTA_IPL_Props )

class GTA_MAP_Props(bpy.types.PropertyGroup):
	@classmethod
	def register( GTA_MAP_Props ):
		GTA_MAP_Props.gta_path = StringProperty(
			name="GTA SA Folder",
			description="Full Path Name of \"GTA San Andreas\"",
			default = "")
		
		GTA_MAP_Props.gtadat_path = StringProperty(
			name="GTA.DAT File Path",
			description="Full Path Name of GTA.DAT",
			default = "")
		
		GTA_MAP_Props.extract_txd = BoolProperty(
			name="Extract TXD",
			description=".",
			default = False )
		
		GTA_MAP_Props.tex_path = StringProperty(
			name="Path",
			description=".",
			default = "")
		
		GTA_MAP_Props.skip_lod = BoolProperty(
			name="Skip LOD objs",
			description=".",
			default = True)
		
		GTA_MAP_Props.skip_nonlod = BoolProperty(
			name="Skip Non-LOD objs",
			description=".",
			default = False)
		
		GTA_MAP_Props.skip_nodes = BoolProperty(
			name="Skip Nodes",
			description=".",
			default = True )
		
		GTA_MAP_Props.skip_binipl = BoolProperty(
			name="Skip BIN-IPLs",
			description=".",
			default = False )
		
		GTA_MAP_Props.alpha_mode = EnumProperty(
			name="Alpha Texture",
			items = (
				("BMP24", "COL+ALP", "Use Grayscale Image File for Alpha Channel"),
				("BMP32", "RGBA",    "Use 32Bit-RGBA Bitmap File") ),
			description="Alpha Image Type for Textures",
			default = "BMP24")
		
		GTA_MAP_Props.active_ipl_id = IntProperty(
			name="active_ipl_id",
			description="Index of the active IPL Entriy",
			default = -1,
			min = -1)
		
		GTA_MAP_Props.ipls = CollectionProperty(
			type = GTA_IPL_Props,
			name = "IPL Entries",
			description = "" )
		
	def ipls_clear( self ):
		self.active_ipl_id = -1
		for ia in range( len( self.ipls ) ): self.ipls.remove(0)

bpy.utils.register_class( GTA_MAP_Props )


# Operators

class OperatorGetIPLs( bpy.types.Operator ):
	bl_idname = "import_ipl.get_ipl"
	bl_label = "Select \"GTA San Andreas\" Folder"
	
	filepath = bpy.props.StringProperty( subtype = "DIR_PATH" )
	filter_glob = StringProperty( default = "", options = {'HIDDEN'} )
	
	def execute( self, context ):
		from . import import_ipl
		import_ipl.GetIPLs( self.filepath )
		return {'FINISHED'}
	
	def invoke( self, context, event ):
		gta_tools = bpy.context.scene.gta_tools
		self.filepath = gta_tools.map_props.gta_path + "\\"
		context.window_manager.fileselect_add( self )
		return {'RUNNING_MODAL'}

class OperatorImportIPL( bpy.types.Operator ):
	bl_idname = "import_ipl.imp_ipl"
	bl_label = "Select IPL file (*.ipl)"
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.set_time_ini()
		gta_tools.dff_props.map = True
		
		from . import import_ipl
		import_ipl.ImportIPL()
		gta_tools.show_msg_elapsed()
		return {'FINISHED'}

class OperatorSetTexPath( bpy.types.Operator ):
	bl_idname = "import_ipl.set_tex_path"
	bl_label = "Select a Folder for Extracting Textures"
	
	filepath = bpy.props.StringProperty( subtype = "DIR_PATH" )
	filter_glob = StringProperty( default = "", options = {'HIDDEN'} )
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.map_props.tex_path = self.filepath.rstrip("\\").lower()
		return {'FINISHED'}
	
	def invoke( self, context, event ):
		gta_tools = bpy.context.scene.gta_tools
		self.filepath = gta_tools.map_props.tex_path + "\\"
		context.window_manager.fileselect_add( self )
		return {'RUNNING_MODAL'}

class OperatorExtTexSelected( bpy.types.Operator ):
	bl_idname = "import_ipl.ext_tex"
	bl_label = "Extract Texrures in Selected Objects"
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.set_time_ini()
		
		from . import import_ipl
		import_ipl.ExtTexSelected()
		gta_tools.show_msg_elapsed()
		return {'FINISHED'}




# UI Panel

class GTA_MAPIO_UI( bpy.types.Panel ):
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'TOOLS'
	bl_context = "objectmode"
	bl_label = "GTA MAP Tools"
	
	def draw( self, context ):
		layout = self.layout
		
		if bpy.context.mode == 'OBJECT':
			gta_tools = bpy.context.scene.gta_tools
			col = layout.column()
			col.row().label( text = "GTA SA Folder:" )
			if "" == gta_tools.map_props.gta_path:
				str_gta_path = "Select a Folder"
			else:
				str_gta_path = gta_tools.map_props.gta_path
			col.row().operator( "import_ipl.get_ipl", str_gta_path )
			is_gta_path_selected = ( "" != gta_tools.map_props.gta_path )
			
			col = layout.column()
			col.enabled = is_gta_path_selected
			col.row().label( text = "IPLs ( Num INSTs ):" )
			col.row().template_list( gta_tools.map_props, "ipls", gta_tools.map_props, "active_ipl_id" )
			is_ipl_selected = ( -1 != gta_tools.map_props.active_ipl_id ) and is_gta_path_selected
			
			col = layout.column()
			col.enabled = is_ipl_selected
			if is_ipl_selected:
				str_load = "Load: " + gta_tools.map_props.ipls[gta_tools.map_props.active_ipl_id].ipl_name
			else:
				str_load = "Load: "
			col.row().operator( "import_ipl.imp_ipl", text = str_load )
			col.row().prop( gta_tools.map_props, "skip_nonlod" )
			col.row().prop( gta_tools.map_props, "skip_lod" )
			col.row().prop( gta_tools.map_props, "skip_nodes" )
			col.row().prop( gta_tools.map_props, "skip_binipl" )
			
			
			col = layout.column()
			col.row().prop( gta_tools.map_props, "alpha_mode" )
			
			col.row().label( text = "TXD Extract:" )
			box = col.box()
			boxcol = box.column( align = False )
			
			row = boxcol.row( align = True )
			row.label( text = "Extract Folder:" )
			
			row = boxcol.row( align = True )
			if "" == gta_tools.map_props.tex_path:
				str_tex_path = "Select a Folder"
			else:
				str_tex_path = gta_tools.map_props.tex_path
			row.operator( "import_ipl.set_tex_path", str_tex_path )
			is_texpath_selected = ( "" != gta_tools.map_props.tex_path )
			
			row = boxcol.row( align = True )
			row.operator( "import_ipl.ext_tex", "Extact TXD @SelObjs" )
			row.enabled = is_texpath_selected
			



### IFP Import/Export ###
# Data Class

class GTA_IFP_AnimProps(bpy.types.PropertyGroup):
	@classmethod
	def register( GTA_IFP_AnimProps ):
		from bpy.props import PointerProperty, StringProperty, BoolProperty, EnumProperty, IntProperty, CollectionProperty
		
		GTA_IFP_AnimProps.name = StringProperty(
			name="Anim Name",
			description="",
			default = "")

bpy.utils.register_class( GTA_IFP_AnimProps )

class GTA_IFP_Props(bpy.types.PropertyGroup):
	@classmethod
	def register( GTA_IFP_Props ):
		from bpy.props import PointerProperty, StringProperty, BoolProperty, EnumProperty, IntProperty, CollectionProperty
		
		## for import
		GTA_IFP_Props.ui_import = BoolProperty(
			name="Import IFP:",
			description="Show UI for Import IFP",
			default = False)
		
		GTA_IFP_Props.filepath = StringProperty(
			name="IFP File Path",
			description="Full Path Name of target IFP",
			default = "")
		
		GTA_IFP_Props.ifp_name = StringProperty(
			name="IFP Name",
			description="Internal IFP Name",
			default = "")
		
		## import options
		GTA_IFP_Props.reset_anim = BoolProperty(
			name="Reset Anim",
			description="Clear All Anim Data in Selected Armature, before importing.",
			default = True)
		
		### Not Useed ###
		#GTA_IFP_Props.reset_arm = BoolProperty(
		#	name="Reset Armature",
		#	description="Clear POS/ROT of Selected Armature, when import with \"Reset Anim\" option.",
		#	default = True)
		
		GTA_IFP_Props.skip_pos = BoolProperty(
			name="Skip POS Keys",
			description="Skip Pos Keys in Selected Directions",
			default = True)
		
		GTA_IFP_Props.skip_root = BoolVectorProperty(
			name="Root",
			description="Skip Root Bone's Pos Keys in Selected Directions",
			default=( False, False, False ),
			subtype='XYZ')
		
		GTA_IFP_Props.skip_children = BoolProperty(
			name="Child Bones",
			description="Skip Child Bone's Pos Keys",
			default = True)
		
		GTA_IFP_Props.use_current_root = BoolProperty(
			name="Use Current Root",
			description="Set Anim based on Current Location of RootBone in Selected Axises",
			default = False)
		
		GTA_IFP_Props.use_current_root_pos = BoolVectorProperty(
			name="POS",
			description="Set Anim based on Current Location of RootBone in Selected Axises",
			#default=( False, False, False ),
			default=( True, True, True ),
			subtype='XYZ')
		
		GTA_IFP_Props.use_current_root_rot = EnumProperty(
			name="ROT",
			items = (
				("NONE",   "None", ""),
				("ALL",    "All", ""),
				("XY",     "XY", ""),
				("YZ",     "YZ", ""),
				("ZX",     "ZX", "") ),
			description="Set Anim based on Current Rotation of RootBone in Selected Plane",
			default = "NONE")
		
		### Not Used ###
		#GTA_IFP_Props.use_current_root_ofs = BoolProperty(
		#	name="Offset Init Value",
		#	description="Set Anim based on Current Location of RootBone in Selected Axises",
		#	default = True)
		
		GTA_IFP_Props.root_to_arm = BoolProperty(
			name="Root to Armature",
			description="Set Root Keys to Armature.",
			default=False)
		
		GTA_IFP_Props.root_to_arm_pos = BoolVectorProperty(
			name="POS",
			description="Set Root POS Keys to Armature.",
			#default=( False, False, False ),
			default=( True, True, True ),
			subtype='XYZ')
		
		GTA_IFP_Props.root_to_arm_rot = EnumProperty(
			name="ROT",
			items = (
				("NONE",   "None", ""),
				("ALL",    "All", ""),
				("XY",     "XY", ""),
				("YZ",     "YZ", ""),
				("ZX",     "ZX", "") ),
			description="Set Root POT Keys to Armature in Selected Plane.",
			default = "NONE")
		
		### Not Used ###
		#GTA_IFP_Props.auto_connect_key = BoolProperty(
		#	name="Auto Connenct Key",
		#	description="Set Keys at the Latest Flame and the Pre-Latest Frame automatically. Connect-key Setting Options are bellow. ",
		#	default=False)
		
		### Not Used ###
		#GTA_IFP_Props.show_connect_options = BoolProperty(
		#	name="Show Options",
		#	description="Show Connect-key Setting Options",
		#	default=False)
		
		### Not Used ###
		#GTA_IFP_Props.set_connect_keys_pos = BoolVectorProperty(
		#	name="POS",
		#	description="Set Keys at Current Flame and Previous Frame",
		#	default=( False, False, False ),
		#	subtype='XYZ')
		
		### Not Used ###
		#GTA_IFP_Props.set_connect_keys_pos_swap = BoolVectorProperty(
		#	name="POS",
		#	description="Set Keys at Current Flame and Previous Frame",
		#	default=( False, False, False ),
		#	subtype='XYZ')
		
		### Not Used ###
		#GTA_IFP_Props.set_connect_keys_rot = BoolVectorProperty(
		#	name="ROT",
		#	description="Set Keys at Current Flame and Previous Frame",
		#	default=( False, ),
		#	size = 1 ,
		#	subtype='NONE')
		
		### Not Used ###
		#GTA_IFP_Props.set_connect_keys_rot_swap = EnumProperty(
		#	name="ROT",
		#	items = (
		#		("NONE",   "None", ""),
		#		("ALL",    "All", ""),
		#		("XY",     "XY", ""),
		#		("YZ",     "YZ", ""),
		#		("ZX",     "ZX", "") ),
		#	description="Set Keys at Current Flame and Previous Frame",
		#	default = "NONE")
		
		### Not Used ###
		#GTA_IFP_Props.overwrite_mode = EnumProperty(
		#	items = (
		#		("ALL",        "All Bones", ""),
		#		("CONFLICTED", "Conflicted Bones", "") ),
		#	description="Set Root POT Keys to Armature in Selected Plane.",
		#	default = "CONFLICTED")
		
		### Not Used ###
		#GTA_IFP_Props.linear_interpolation = BoolProperty(
		#	name="Linear Interpolation",
		#	description="Set Linear Interpolation to Existing Previous Frame, to Avoid \"Overshoot\" by Bezier-Interpolation",
		#	default = True)
		
		### Not Used ###
		#GTA_IFP_Props.clear_following = BoolProperty(
		#	name="Clear Follow Frames",
		#	description="Clear Following Frames",
		#	default = False)
		
		GTA_IFP_Props.show_frame_ops_imp = BoolProperty(
			name="Show Frame Options",
			description="Show Frame Options (Option-Settings are effective, if this menu is hidden)",
			default = False)
		
		GTA_IFP_Props.show_frame_ops_exp = BoolProperty(
			name="Show Frame Options",
			description="Show Frame Options (Option-Settings are effective, if this menu is hidden)",
			default = False)
		
		GTA_IFP_Props.auto_snap = BoolProperty(
			name="Snap Time Keys",
			description="Snap Time Keys to the Nealest Flame",
			default = False)
		
		GTA_IFP_Props.frame_rate_preset = EnumProperty(
			items = (
				("30FPS" , "30fps" , ""),
				("60FPS" , "60fps" , ""),
				("CUSTOM", "Custom", ""),
				("RENDER", "Render", "") ),
			description="Frame Rate used for mapping IFP Frames to Blender Animation. (\"Render\" : Use Render Frame Rate of Blender Settings)",
			default = "30FPS")
		
		GTA_IFP_Props.frame_rate = IntProperty(
			name="fps",
			description="Custom Frame Rate.",
			min = 10, max = 120, default = 30)
		
		GTA_IFP_Props.adjust_render_rate = BoolProperty(
			name="Adjust Render F.R.",
			description="Set Render Frame Rate of Blender, as the value set here.",
			default = True)
		
		GTA_IFP_Props.adjust_scene_range = BoolProperty(
			name="Adjust Scene Range",
			description="Set Scene Frame Range, as Imported Anim.",
			default = True)
		
		GTA_IFP_Props.load_at_end_anim = BoolProperty(
			name="Load at End Time",
			description = "Load Anim at Time-Based End Key of Current Animation Data.( use in a case that the End of the Current Data is not snapped to any frames )",
			default = False)
		
		GTA_IFP_Props.clear_pose_selbone = BoolProperty(
			name="Clear Pose",
			description="Clear Pose of Selected Bones",
			default = True)
		
		GTA_IFP_Props.set_inirot_selbone = BoolProperty(
			name="Set Init-Rot-Key",
			description="Set Rotation Key at Initial Frame",
			default = True)
		
		GTA_IFP_Props.ui_utilities = BoolProperty(
			name="Utilities",
			description="IFP Utilities",
			default = False)
		
		GTA_IFP_Props.reset_selbone_ops = BoolProperty(
			name="Reset Selected Bones",
			description="Reset Selected Bones",
			default = False)
		
		GTA_IFP_Props.active_anim_id = IntProperty(
			name="active_anim_id",
			description="Index of the active IFP Anim Name",
			default = -1,
			min = -1)
		
		GTA_IFP_Props.anims = CollectionProperty(
			type=GTA_IFP_AnimProps,
			name="IFP Anims",
			description="")
		
		## for export
		GTA_IFP_Props.ui_export = BoolProperty(
			name="Export IFP:",
			description="Show UI for Export IFP",
			default = False)
		
		GTA_IFP_Props.exp_filepath = StringProperty(
			name="IFP File Path",
			description="Full Path Name of target IFP",
			default = "")
		
		GTA_IFP_Props.exp_internal_file_name = StringProperty(
			name="",
			description="Internal IFP file Name",
			default = "")
		
		GTA_IFP_Props.exp_anim_name = StringProperty(
			name="",
			description="Anim Name for Export",
			default = "")
		
		GTA_IFP_Props.ifp_format = EnumProperty(
			items = (
				("ANP3", "ANP3", ""),
				("ANPK", "ANPK", "") ),
			description="IFP Data Format ( \"ANP3\" for ActionScene / \"ANPK\" for CutScene )",
			default = "ANP3")
		
		GTA_IFP_Props.rev_bone = BoolProperty(
			name="Original Bone Name:",
			description="Use Original Bone Name of the Loaded Character Model (for internal description of IFP)",
			default = True)
		
		
	def anims_clear( self ):
		self.active_anim_id = -1
		for ia in range( len( self.anims ) ): self.anims.remove(0)
	
	
bpy.utils.register_class( GTA_IFP_Props )


# Operators

class OperatorSelectIFP( bpy.types.Operator ):
	bl_idname = "import_ifp.sel_ifp"
	bl_label = "Select IFP file (*.ifp)"
	
	filepath = bpy.props.StringProperty( subtype = "FILE_PATH" )
	filename_ext = ".ifp"
	filter_glob = StringProperty( default = "*.ifp", options = {'HIDDEN'} )
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.ifp_props.filepath = self.filepath
		gta_tools.init_msg( "--- Amin List Read ---" )
		
		from . import import_ifp
		import_ifp.ImportIFP( 0x0 )  # flags(0x1):  False: read list, True: read active anim
		gta_tools.show_msg( True, True )
		return {'FINISHED'}
	
	def invoke( self, context, event ):
		gta_tools = bpy.context.scene.gta_tools
		self.filepath = gta_tools.ifp_props.filepath
		context.window_manager.fileselect_add( self )
		return {'RUNNING_MODAL'}

class OperatorImportAnim( bpy.types.Operator ):
	bl_idname = "import_ifp.imp_anim"
	bl_label = "Import Anim"
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.init_msg( "--- IFP Import ---" )
		
		from . import import_ifp
		import_ifp.ImportIFP( 0x1 )  # flags(0x1):  False: read list, True: read active anim
		gta_tools.show_msg( False, True )
		return {'FINISHED'}

class OperatorExportIFP( bpy.types.Operator ):
	bl_idname = "export_ifp.exp_ifp"
	bl_label = "Select IFP file (*.ifp)"
	
	filepath = bpy.props.StringProperty( subtype = "FILE_PATH" )
	filename_ext = ".ifp"
	filter_glob = StringProperty( default = "*.ifp", options = {'HIDDEN'} )
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.init_msg( "--- IFP Export ---" )
		gta_tools.ifp_props.exp_filepath = self.filepath
		
		from . import export_ifp
		export_ifp.ExportIFP()
		gta_tools.show_msg( True, True )
		return {'FINISHED'}
	
	def invoke( self, context, event ):
		gta_tools = bpy.context.scene.gta_tools
		self.filepath = gta_tools.ifp_props.exp_filepath
		context.window_manager.fileselect_add( self )
		return {'RUNNING_MODAL'}

class OperatorResetAnim( bpy.types.Operator ):
	bl_idname = "import_ifp.reset_anim"
	bl_label = "Reset Anim"
	
	def execute( self, context ):
		from . import import_ifp
		import_ifp.ResetAnim()
		import_ifp.ResetPose()
		return {'FINISHED'}

class OperatorAnimDirection( bpy.types.Operator ):
	bl_idname = "import_ifp.anim_direction"
	bl_label = "Set Anim Direction"
	
	def execute( self, context ):
		from . import import_ifp
		import_ifp.AnimDirection()
		return {'FINISHED'}

class OperatorResetArmature( bpy.types.Operator ):
	bl_idname = "import_ifp.reset_armature"
	bl_label = "Reset Armature's POS/ROT"
	
	def execute( self, context ):
		from . import import_ifp
		import_ifp.ResetArmature()
		return {'FINISHED'}

class OperatorSetConnectKeys( bpy.types.Operator ):
	bl_idname = "import_ifp.set_connect_keys"
	bl_label = "Set Connect Keys"
	
	def execute( self, context ):
		from . import import_ifp
		import_ifp.SetConnectKeys()
		return {'FINISHED'}

class OperatorSelectKeyedBones( bpy.types.Operator ):
	bl_idname = "import_ifp.sel_keyed_bones"
	bl_label = "Select Keyed Bones"
	
	def execute( self, context ):
		from . import import_ifp
		import_ifp.SelectKeyedBones()
		return {'FINISHED'}

class OperatorResetSelectedBones( bpy.types.Operator ):
	bl_idname = "import_ifp.reset_sel_bones"
	bl_label = "Reset Selected Bones"
	
	def execute( self, context ):
		from . import import_ifp
		import_ifp.ResetSelectedBones()
		return {'FINISHED'}


# UI Panel

class GTA_IFPIO_UI( bpy.types.Panel ):
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'TOOLS'
	bl_context = "posemode"
	bl_label = "GTA IFP Tools"
	
	def draw( self, context ):
		layout = self.layout
		
		if bpy.context.mode == 'POSE':
			gta_tools = bpy.context.scene.gta_tools
			col = layout.column( align = True )
			col.label( text = "- - - - - - - - - -" )
			col.prop( gta_tools.ifp_props, "ui_import" )
			if gta_tools.ifp_props.ui_import:
				is_ifp_selected = ( "" != gta_tools.ifp_props.ifp_name )
				if is_ifp_selected:
					col.operator( "import_ifp.sel_ifp", text = "File : " + os.path.split( gta_tools.ifp_props.filepath )[1] )
				else:
					col.operator( "import_ifp.sel_ifp", text = "Select File" )
				row_animlist = col.row()
				row_animlist.template_list( gta_tools.ifp_props, "anims", gta_tools.ifp_props, "active_anim_id" )
				row_animlist.enabled = is_ifp_selected
				is_anim_selected = ( -1 != gta_tools.ifp_props.active_anim_id )
				
				row_load = col.row()
				row_load.enabled = is_anim_selected
				if is_anim_selected:
					str_load = "Load : " + gta_tools.ifp_props.anims[gta_tools.ifp_props.active_anim_id].name
				else:
					str_load = "Load :"
				row_load.operator( "import_ifp.imp_anim", text = str_load )
				
				col = layout.column( align = True )
				col.prop( gta_tools.ifp_props, "reset_anim" )
				
				col.prop( gta_tools.ifp_props, "skip_pos" )
				if gta_tools.ifp_props.skip_pos:
					boxcol = col.box().column( align = True )
					boxcol.row().prop( gta_tools.ifp_props, "skip_root" )
					boxcol.prop( gta_tools.ifp_props, "skip_children" )
					col = layout.column( align = True )
				
				col.prop( gta_tools.ifp_props, "use_current_root" )
				if gta_tools.ifp_props.use_current_root:
					boxcol = col.box().column( align = True )
					boxcol.row().prop( gta_tools.ifp_props, "use_current_root_pos" )
					boxcol.prop( gta_tools.ifp_props, "use_current_root_rot" )
					col = layout.column( align = True )
				
				col.prop( gta_tools.ifp_props, "root_to_arm" )
				if gta_tools.ifp_props.root_to_arm:
					boxcol = col.box().column( align = True )
					boxcol.row().prop( gta_tools.ifp_props, "root_to_arm_pos" )
					boxcol.prop( gta_tools.ifp_props, "root_to_arm_rot" )
					#boxcol.prop( gta_tools.ifp_props, "reset_arm" )
					col = layout.column( align = True )
				
				col.prop( gta_tools.ifp_props, "show_frame_ops_imp" )
				if gta_tools.ifp_props.show_frame_ops_imp:
					boxcol = col.box().column( align = True )
					boxcol.label( "Frame Rate:" )
					boxcol.prop( gta_tools.ifp_props, "frame_rate_preset" )
					if "CUSTOM" == gta_tools.ifp_props.frame_rate_preset:
						boxcol.prop( gta_tools.ifp_props, "frame_rate" )
					if "RENDER" != gta_tools.ifp_props.frame_rate_preset:
						boxcol.prop( gta_tools.ifp_props, "adjust_render_rate" )
					boxcol.prop( gta_tools.ifp_props, "adjust_scene_range" )
					boxcol.prop( gta_tools.ifp_props, "auto_snap" )
					boxcol.prop( gta_tools.ifp_props, "load_at_end_anim" )
					col = layout.column( align = True )
				
				## !!! under construction
				#col.row().prop( gta_tools.ifp_props, "auto_connect_key" )
				#if gta_tools.ifp_props.auto_connect_key:
				#	box = col.box()
				#	boxcol = box.column()
				#	boxcol.row().prop( gta_tools.ifp_props, "reset_arm" )
				
				## !!! under construction
				#col.row().label( "OverWriteMode:" )
				#col.row().prop( gta_tools.ifp_props, "overwrite_mode" )
				
				#col.row().label( text = "" )
			
			col = layout.column( align = True )
			col.label( text = "- - - - - - - - - -" )
			col.prop( gta_tools.ifp_props, "ui_export" )
			if gta_tools.ifp_props.ui_export:
				col.operator( "export_ifp.exp_ifp", text = "Export" )
				row = col.row().split( 0.4 )
				row.label( "Format:" )
				row.prop( gta_tools.ifp_props, "ifp_format" )
				col.label( "IFP/Anim Name:" )
				boxcol = col.box().column( align = True )
				boxcol.label( "Internal File Name:" )
				boxcol.prop( gta_tools.ifp_props, "exp_internal_file_name" )
				boxcol.label( "Animation Name:" )
				boxcol.prop( gta_tools.ifp_props, "exp_anim_name" )
				col = layout.column( align = True )
				#col.prop( gta_tools.ifp_props, "rev_bone" )
				
				col.prop( gta_tools.ifp_props, "show_frame_ops_exp" )
				if gta_tools.ifp_props.show_frame_ops_exp:
					boxcol = col.box().column( align = True )
					boxcol.label( "Frame Rate:" )
					boxcol.prop( gta_tools.ifp_props, "frame_rate_preset" )
					if "CUSTOM" == gta_tools.ifp_props.frame_rate_preset:
						boxcol.prop( gta_tools.ifp_props, "frame_rate" )
					col = layout.column( align = True )
			
			col = layout.column( align = True )
			col.label( text = "- - - - - - - - - -" )
			col.prop( gta_tools.ifp_props, "ui_utilities" )
			if gta_tools.ifp_props.ui_utilities:
				col = layout.column( align = True )
				col.label( "Animation:" )
				col.operator( "import_ifp.reset_anim", text = "Reset Anim" )
				col.label( "Armature:" )
				col.operator( "import_ifp.reset_armature", text = "Reset Armature" )
				col.label( "Bones:" )
				col.operator( "import_ifp.anim_direction", text = "Anim Direction" )
				col.operator( "import_ifp.sel_keyed_bones", text = "Select Keyed Bones" )
				col = layout.column( align = True )
				boxcol = col.box().column( align = True )
				boxcol.operator( "import_ifp.reset_sel_bones", text = "Reset Seleced Bones" )
				boxcol.prop( gta_tools.ifp_props, "clear_pose_selbone" )
				boxcol.prop( gta_tools.ifp_props, "set_inirot_selbone" )
				col.label( "System:" )
				col.operator( "gta_utils.init_ifp_tools", text = "Init IFP Tools" )
			
			## !!! under construction
			#box = col.box()
			#boxcol = box.column()
			#boxcol.row().operator( "import_ifp.set_connect_keys", text = "Set Connect Keys" )
			#boxcol.row().prop( gta_tools.ifp_props, "show_connect_options" )
			#if gta_tools.ifp_props.show_connect_options:
			#	boxcol.row().label( "Set Axis:" )
			#	boxcol.row().prop( gta_tools.ifp_props, "set_connect_keys_pos" )
			#	boxcol.row().prop( gta_tools.ifp_props, "set_connect_keys_rot" )
			#	boxcol.row().label( "Swap Root and Armature:" )
			#	boxcol.row().prop( gta_tools.ifp_props, "set_connect_keys_pos_swap" )
			#	boxcol.row().prop( gta_tools.ifp_props, "set_connect_keys_rot_swap" )
			#	boxcol.row().prop( gta_tools.ifp_props, "linear_interpolation" )
			#	boxcol.row().prop( gta_tools.ifp_props, "clear_following" )




### Weight Tools ###
# Data Class

class GTA_WEIGHT_Props(bpy.types.PropertyGroup):
	@classmethod
	def register( GTA_WEIGHT_Props ):
		GTA_WEIGHT_Props.cur_loc_1st = FloatVectorProperty(
			name="cur_loc_1st",
			description="" )
		
		GTA_WEIGHT_Props.cur_loc_2nd = FloatVectorProperty(
			name="cur_loc_2nd",
			description="" )
		
		GTA_WEIGHT_Props.grad_contour = EnumProperty(
			items = (
				("LINE", "Line", ""),
				("CIRCLE", "Circle", "") ),
			name="Contour",
			description="",
			default = "LINE")
		
		GTA_WEIGHT_Props.grad_and_norm = BoolProperty(
			name="Normalize",
			description="Normalize with Grad Operation",
			default = True)
		
		GTA_WEIGHT_Props.grad_range = FloatVectorProperty(
			name="range",
			description="",
			size = 2,
			default = ( 0.0, 1.0 ),
			max = 1.0,
			min = 0.0 )
			
		GTA_WEIGHT_Props.sel_type = EnumProperty(
			name="target",
			items = (
				("NONE", "No Groups", ""),
				("OVER4", "> 4 Groups", "") ),
			description="",
			default = "NONE")
		
		GTA_WEIGHT_Props.grad_flags = IntProperty(
			name="Weight Grad Status",
			description="Weight Grad Status",
			default = 0x0 )
		


bpy.utils.register_class( GTA_WEIGHT_Props )


# Operators

class OperatorVWGradSet1st( bpy.types.Operator ):
	bl_idname = "weight_tools.vw_set_1st"
	bl_label = "test"
	
	def execute( self, context ):
		from . import weight_tools
		weight_tools.vw_set_1st()
		return {'FINISHED'}

class OperatorVWGradSet2nd( bpy.types.Operator ):
	bl_idname = "weight_tools.vw_set_2nd"
	bl_label = "--- Weight Gradient ---"
	
	def execute( self, context ):
		from . import weight_tools
		weight_tools.vw_set_2nd()
		return {'FINISHED'}


class OperatorVWGrad( bpy.types.Operator ):
	bl_idname = "weight_tools.vw_grad"
	bl_label = "test"
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.set_time_ini()
		from . import weight_tools
		weight_tools.vw_grad()
		gta_tools.show_msg_elapsed()
		return {'FINISHED'}

class OperatorVWGradClear( bpy.types.Operator ):
	bl_idname = "weight_tools.vw_grad_clear"
	bl_label = "test"
	
	def execute( self, context ):
		from . import weight_tools
		weight_tools.vw_grad_clear()
		return {'FINISHED'}

class OperatorVWClear( bpy.types.Operator ):
	bl_idname = "weight_tools.vw_crear"
	bl_label = "test"
	
	def execute( self, context ):
		from . import weight_tools
		weight_tools.vw_clear()
		return {'FINISHED'}

class OperatorVWClearZero( bpy.types.Operator ):
	bl_idname = "weight_tools.vw_crear_zero"
	bl_label = "test"
	
	def execute( self, context ):
		from . import weight_tools
		weight_tools.vw_clear_zero()
		return {'FINISHED'}

class OperatorVWNormalize( bpy.types.Operator ):
	bl_idname = "weight_tools.vw_normalize"
	bl_label = "test"
	
	def execute( self, context ):
		from . import weight_tools
		weight_tools.vw_normalize()
		return {'FINISHED'}

class OperatorVWSelect( bpy.types.Operator ):
	bl_idname = "weight_tools.vw_select"
	bl_label = "test"
	
	def execute( self, context ):
		from . import weight_tools
		weight_tools.vw_select()
		return {'FINISHED'}

class SetupDrawCallBack(bpy.types.Operator):
	bl_idname = "weight_tools.setup_draw_callback"
	bl_label = "SetUp Draw Callback"

	def execute(self, context):
		callbacks = weight_tools.callbacks
		region = context.region
		region_id = hash(region)
		cb_data = callbacks.get(region_id)
		if cb_data is None:
			hcb_px   = region.callback_add(weight_tools.draw_callback_px  , (self, context), 'POST_PIXEL')
			hcb_view = region.callback_add(weight_tools.draw_callback_view, (self, context), 'POST_VIEW')
			callbacks[region_id] = region, hcb_px, hcb_view
			print( "Enabled Draw CallBack" )
		else:
			region.callback_remove(cb_data[1])
			region.callback_remove(cb_data[2])
			del callbacks[region_id]
			print( "Disabled Draw CallBack" )
		
		context.area.tag_redraw()
		return {'FINISHED'}


# UI Panel

class GTA_VWTOOL_UI( bpy.types.Panel ):
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'TOOLS'
	bl_context = "mesh_edit"
	bl_label = "GTA Weight Tools"
	
	def draw( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		
		layout = self.layout
		col = layout.column( align = True )
		col.label( text = "Clear/Normalize:" )
		row = col.row()
		row.operator( "weight_tools.vw_crear_zero", "Clear Zero" )
		row.operator( "weight_tools.vw_normalize", "Normalize" )
		row = col.row()
		row.operator( "weight_tools.vw_crear", "Clear All" )
		
		col = layout.column( align = True )
		col = layout.column( align = False )
		col.label( text = "Select Verts:" )
		col.row().prop( gta_tools.weight_props, "sel_type" )
		col.row().operator( "weight_tools.vw_select", "Select" )
		
		col = layout.column( align = True )
		col = layout.column( align = True )
		col.label( text = "Gradient:" )
		col.row().prop( gta_tools.weight_props, "grad_range" )
		col = layout.column( align = True )
		col.row().prop( gta_tools.weight_props, "grad_contour" )
		row = col.row()
		col.row().prop( gta_tools.weight_props, "grad_and_norm" )
		
		col = layout.column( align = True )
		row = col.row()
		row.operator( "weight_tools.vw_set_1st", "Set 1st" )
		row.operator( "weight_tools.vw_set_2nd", "Set 2nd" )
		row.operator( "weight_tools.vw_grad_clear", "Cancel" )
		#row.enabled = gta_tools.weight_props.grad_flags & weight_tools.GRAD_SET_1ST
		col = layout.column( align = True )
		row = col.row()
		row.operator( "weight_tools.vw_grad", "Grad" )
		row.enabled = ( gta_tools.weight_props.grad_flags & weight_tools.GRAD_SET_1ST ) and ( gta_tools.weight_props.grad_flags & weight_tools.GRAD_SET_2ND )
		col = layout.column( align = True )
		#row = col.row()
		#row.operator( "weight_tools.setup_draw_callback", "Test Code #01" )



### GTA Tools Utility
# Data Class

class GTA_UTIL_Props(bpy.types.PropertyGroup):
	@classmethod
	def register( GTA_UTIL_Props ):
		GTA_UTIL_Props.target_all = BoolProperty(
			name="ALL",
			description="Collisions/Shadows",
			default = True)
		
		GTA_UTIL_Props.target_coll = BoolProperty(
			name="COLL",
			description="Collisions/Shadows",
			default = True)
		
		GTA_UTIL_Props.target_vlo = BoolProperty(
			name="VLO",
			description="\"VLO\" Objs",
			default = True)
		
		GTA_UTIL_Props.target_ok = BoolProperty(
			name="OK",
			description="\"OK\" Objs",
			default = False)
		
		GTA_UTIL_Props.target_dam = BoolProperty(
			name="DAM",
			description="\"DAM\" Objs",
			default = True)
		
		GTA_UTIL_Props.normal_map = EnumProperty(
			name="Normal Map",
			items = (
				("SET",     "Set",     "Set \"Use Normal Map\""),
				("DESET",   "Deset",   "DeSet \"Use Normal Map\""),
				("CURRENT", "Not Change", "Use Current Setting") ),
			description="Normal Map",
			default = "CURRENT")
		
		GTA_UTIL_Props.normal_factor = FloatProperty(
			name="Normal Factor",
			description="Amount texture affects normal values.",
			default = 1.0)
		
		GTA_UTIL_Props.use_alpha = EnumProperty(
			name="Use Alpha",
			items = (
				("SET",     "Set",     "Set \"Use Alpha\""),
				("DESET",   "Deset",   "DeSet \"Use Alpha\""),
				("CURRENT", "Not Change", "Use Current Setting") ),
			description="Use Alpha",
			default = "CURRENT")
		
		GTA_UTIL_Props.use_transparent_shadows = EnumProperty(
			name="Tra-Shadows",
			items = (
				("SET",     "Set",     "Set \"Receive Transparent Shadows\""),
				("DESET",   "Deset",   "DeSet \"Receive Transparent Shadows\""),
				("CURRENT", "Not Change", "Use Current Setting") ),
			description="Receive Transparent Shadows",
			default = "CURRENT")
		
		GTA_UTIL_Props.transparency_method = EnumProperty(
			name="Tra-Method",
			items = (
				("MASK",             "Mask",             "Mask the background."),
				("Z_TRANSPARENCY",   "Z Transparency",   "Use alpha buffer for transparent faces."),
				("RAYTRACE",         "Raytrace",         "Use raytracing for transparent refraction rendering."),
				("CURRENT", "Not Change", "Use Current Setting") ),
			description="Method to use for rendering transparency",
			default = "CURRENT")

bpy.utils.register_class( GTA_UTIL_Props )


# Operators

class OperatorHideObjSets( bpy.types.Operator ):
	bl_idname = "gta_utils.hide_objsets"
	bl_label = "Set Props for Grouped and Linked Objects"
	val = BoolProperty()
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.set_objsets_prop( 0, self.val )  # mode: 0 = hide/show, 1 = select/deselect
		return {'FINISHED'}

class OperatorSelectObjSets( bpy.types.Operator ):
	bl_idname = "gta_utils.sel_objsets"
	bl_label = "Set Props for Grouped and Linked Objects"
	val = BoolProperty()
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.set_objsets_prop( 1, self.val )  # mode: 0 = hide/show, 1 = select/deselect
		return {'FINISHED'}

class OperatorToggleBoneVisibility( bpy.types.Operator ):
	bl_idname = "gta_utils.tgl_bone"
	bl_label = "Toggle Visibility of Nodes/Bones"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.tgl_bone()
		return {'FINISHED'}

class OperatorToggleNodeVisibility( bpy.types.Operator ):
	bl_idname = "gta_utils.tgl_node"
	bl_label = "Toggle Visibility of Nodes/Bones"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.tgl_node()
		return {'FINISHED'}

class OperatorToggleNameVisibility( bpy.types.Operator ):
	bl_idname = "gta_utils.tgl_name"
	bl_label = "Toggle Visibility of Object Names"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.tgl_name()
		return {'FINISHED'}

class OperatorToggleXRay( bpy.types.Operator ):
	bl_idname = "gta_utils.tgl_xray"
	bl_label = "Toggle X-Ray Propertiy of Nodes"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.tgl_xray()
		return {'FINISHED'}

class OperatorRemoveUnusedData( bpy.types.Operator ):
	bl_idname = "gta_utils.rm_unused"
	bl_label = "Test"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.rm_unused()
		return {'FINISHED'}

class OperatorRemoveImages( bpy.types.Operator ):
	bl_idname = "gta_utils.rm_images"
	bl_label = "Remove Images"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.rm_images()
		return {'FINISHED'}

class OperatorSetPropsSelObj( bpy.types.Operator ):
	bl_idname = "gta_utils.set_props"
	bl_label = "Set Properties in Selected Objects"
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.set_time_ini()
		
		from . import gta_utils
		gta_utils.SetProps()
		gta_tools.show_msg_elapsed()
		return {'FINISHED'}

class OperatorInitGTATools( bpy.types.Operator ):
	bl_idname = "gta_utils.init_gta_tools"
	bl_label = "Initialize Gta Tools Properties"
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.init_msg( "--- Initialize GTA Tools ---" )
		from . import gta_utils
		gta_utils.InitToolProps()
		gta_tools.show_msg( True, True )
		return {'FINISHED'}

class OperatorInitIFPTools( bpy.types.Operator ):
	bl_idname = "gta_utils.init_ifp_tools"
	bl_label = "Initialize IFP Tools Properties"
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.init_msg( "--- Initialize IFP Tools ---" )
		from . import gta_utils
		gta_utils.InitToolProps( ["ifp_props",] )
		gta_tools.show_msg( True, True )
		return {'FINISHED'}

class OperatorInitGTAUtils( bpy.types.Operator ):
	bl_idname = "gta_utils.init_gta_utils"
	bl_label = "Initialize GTA Utility Properties"
	
	def execute( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		gta_tools.init_msg( "--- Initialize GTA Tools Utilities ---" )
		from . import gta_utils
		gta_utils.InitToolProps( ["util_props",] )
		gta_tools.show_msg( True, True )
		return {'FINISHED'}


# Operators

class OperatorTestCode01( bpy.types.Operator ):
	bl_idname = "gta_utils.test01"
	bl_label = "test"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.test01()
		return {'FINISHED'}

class OperatorTestCode02( bpy.types.Operator ):
	bl_idname = "gta_utils.test02"
	bl_label = "test"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.test02()
		return {'FINISHED'}

class OperatorTestCode03( bpy.types.Operator ):
	bl_idname = "gta_utils.test03"
	bl_label = "test"
	
	def execute( self, context ):
		from . import gta_utils
		gta_utils.test03()
		return {'FINISHED'}

# UI Panel

class GTA_UTILS_UI( bpy.types.Panel ):
	bl_space_type = 'VIEW_3D'
	bl_region_type = 'UI'
	bl_label = "GTA Tools Utility"
	
	def draw( self, context ):
		gta_tools = bpy.context.scene.gta_tools
		layout = self.layout
		col = layout.column( align = True )
		#col.label( text = "Grouped Objs Operations:" )
		col.label( text = "Display Toggle in Linked:" )
		
		row = col.row()
		row.operator( "gta_utils.tgl_name", "Name" )
		row.operator( "gta_utils.tgl_node", "Node" )
		row.operator( "gta_utils.tgl_xray", "X-Ray" )
		
		col = layout.column( align = True )
		col.label( text = "Vehicle Objs Operations:" )
		box = col.box()
		boxcol = box.column( align = True )
		row = boxcol.row( align = True )
		row.prop( gta_tools.util_props, "target_all" )
		row = boxcol.row( align = True )
		row.prop( gta_tools.util_props, "target_coll" )
		row.enabled = not gta_tools.util_props.target_all
		row.prop( gta_tools.util_props, "target_vlo" )
		row.enabled  = not gta_tools.util_props.target_all
		row = boxcol.row( align = True )
		row.prop( gta_tools.util_props, "target_ok" )
		row.enabled  = not gta_tools.util_props.target_all
		row.prop( gta_tools.util_props, "target_dam" )
		row.enabled  = not gta_tools.util_props.target_all
		row = boxcol.row( align = True )
		row.operator( "gta_utils.hide_objsets", "Show" ).val = False
		row.operator( "gta_utils.hide_objsets", "Hide" ).val = True
		row = boxcol.row( align = True )
		
		row.operator( "gta_utils.sel_objsets", "Select"   ).val = True
		row.operator( "gta_utils.sel_objsets", "DeSelect" ).val = False
		
		col = layout.column()
		col.row().label( text = "Material/Texture Props:" )
		box = col.box()
		boxcol = box.column( align = False )
		
		row = boxcol.row()
		row.operator( "gta_utils.set_props", "SetProps @SelObjs" )
		row = boxcol.row()
		row.prop( gta_tools.util_props, "normal_map" )
		row = boxcol.row( align = False )
		row.prop( gta_tools.util_props, "normal_factor" )
		row.enabled = ( "SET" == gta_tools.util_props.normal_map )
		boxcol.row().prop( gta_tools.util_props, "use_alpha" )
		boxcol.row().prop( gta_tools.util_props, "use_transparent_shadows" )
		boxcol.row().prop( gta_tools.util_props, "transparency_method" )
		
		col = layout.column( align = True )
		#row = col.row()
		col.label( text = "System Tools" )
		#row = col.row()
		col.operator( "gta_utils.rm_unused", "Remove UnUsed Data" )
		col.operator( "gta_utils.rm_images", "Remove All Images" )
		col.operator( "gta_utils.init_gta_tools", "Init GTA Tools All" )
		col.operator( "gta_utils.init_gta_utils", "Init GTA Tools Utility" )
		
		col = layout.column( align = True )
		row = col.row()
		row.label( text = "Codes for DeBUG:" )
		row = col.row()
		row.operator( "gta_utils.test01", "Test Code #01" )
		row = col.row()
		row.operator( "gta_utils.test02", "Test Code #02" )
		row = col.row()
		row.operator( "gta_utils.test03", "Test Code #03" )



### Generic ###
# Data Class

class GTATOOLS_Props(bpy.types.PropertyGroup):
	@classmethod
	def register( GTATOOLS_Props ):
		from bpy.props import PointerProperty, StringProperty, BoolProperty, EnumProperty, IntProperty, CollectionProperty
		
		GTATOOLS_Props.dff_props = PointerProperty(
			type=GTA_DFF_Props,
			name="DFF Props",
			description="Properties for DFF format @GTA Tools")
		
		GTATOOLS_Props.map_props = PointerProperty(
			type=GTA_MAP_Props,
			name="MAP Props",
			description="Properties for MAP Import @GTA Tools")
		
		GTATOOLS_Props.ifp_props = PointerProperty(
			type=GTA_IFP_Props,
			name="IFP Props",
			description="Properties for IFP format @GTA Tools")
		
		GTATOOLS_Props.weight_props = PointerProperty(
			type=GTA_WEIGHT_Props,
			name="Weight Props",
			description="Properties for Weight Tools @GTA Tools")
		
		GTATOOLS_Props.util_props = PointerProperty(
			type=GTA_UTIL_Props,
			name="Utility Props",
			description="Properties for Utilities @GTA Tools")
		
		GTATOOLS_Props.msg = StringProperty(
			name="Message",
			description="Message",
			default = "")
		
		GTATOOLS_Props.time_ini = FloatProperty(
			name="Initial Time Stump",
			description="Initial Time Stump",
			default = 0.0 )
		
		bpy.types.Scene.gta_tools = PointerProperty(
			type=GTATOOLS_Props,
			name="GTA Tools Props",
			description="GTA Tools Properties")
	
	def set_time_ini( self ):
		self.time_ini = time.clock()
	
	def init_msg( self, init_str = None, init_time = True ):
		self.set_msg( init_str, dump_flg = False, init_flg = True, init_time_flg = True )
	
	def set_msg( self, msg, dump_flg = True, cr_flg = True, init_flg = False, init_time_flg = False ):
		if init_flg:
			self.msg = ""
		if init_time_flg:
			self.time_ini = time.clock()
		self.msg += msg
		if cr_flg: self.msg += "\n"
		if dump_flg: print( msg )
	
	def show_msg( self, show_flg = True, elap_flg = False ):
		if elap_flg:
			str_erap = "Elapsed Time : %.3lf sec" %( ( time.clock() - self.time_ini ) )
			print( str_erap )
			self.msg += str_erap
		if show_flg:
			bpy.ops.scene.msg_popup( 'INVOKE_DEFAULT', msg = self.msg )
	
	def show_msg_elapsed( self ):
		self.msg += "Elapsed Time : %.3lf sec" %( ( time.clock() - self.time_ini ) )
		bpy.ops.scene.msg_popup( 'INVOKE_DEFAULT', msg = self.msg )
		print( self.msg )
	
	def show_msg_elapsed_tmp( self ):
		self.set_msg( "Elapsed Time : %.3lf sec" %( ( time.clock() - self.time_ini ) ), True, False )
		bpy.ops.scene.msg_popup( 'INVOKE_DEFAULT', msg = self.msg )
	
	def reset_tool_props( self, str_props = [] ):
		results = []
		if "gta_tools" in bpy.context.scene.keys():
			for prop in bpy.context.scene.gta_tools.items():
				if prop[0] not in ( "msg", "time_ini" ):
					if 0 == len( str_props ) or prop[0] in str_props:
						bpy.ops.wm.properties_remove(data_path="scene.gta_tools", property = prop[0] )
						results.append( prop[0] )
		return results
	
	@classmethod
	def unregister( GTATOOLS_Props ):
		del bpy.types.Scene.gta_tools

bpy.utils.register_class( GTATOOLS_Props )


# Operators

class MsgPopupOperator(bpy.types.Operator):
	bl_idname = "scene.msg_popup"
	bl_label = "--- Message ---"
	msg = bpy.props.StringProperty()
	
	def execute(self, context):
		return {'FINISHED'}
	
	def invoke(self, context, event):
		wm = context.window_manager
		return wm.invoke_popup(self)
	
	def draw(self, context):
		layout = self.layout
		lines = self.msg.split('\n')
		col = layout.column()
		for line in lines:
			col.label( text = line )



### Test Codes ###
# Data Class


## Registring

def register():
	bpy.utils.register_module( __name__ )
	
def unregister():
	weight_tools.callbacks_clear()
	bpy.utils.unregister_module( __name__ )

if __name__ == "__main__":
	register()
