# extract_txd.py @space_view3d_gta_tools
# 2011 ponz
script_info = "GTA IFP Importer ( built:2012.02.05 )"

# ***** BEGIN GPL LICENSE BLOCK *****
#
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or ( at your option ) any later version.
#
# This program is distributed in the hope that it will be useful, 
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.	See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software Foundation, 
# Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
# ***** END GPL LICENCE BLOCK *****

import bpy
import os
import time
import struct
from bpy.props import *
from mathutils import *
from math import *
from rna_prop_ui import rna_idprop_ui_prop_get
import ctypes
import re
import sys
from zlib import crc32, compress

## Classes
# Data Class
class ClassTXDTexture:
	def __init__( self ):
		self.col_name  = ""
		self.col_path  = ""
		self.col_tex   = None
		self.alp_name  = ""
		self.alp_path  = ""
		self.alp_tex   = None
		self.img_path  = ""
		self.extracted = False

#### dxt decoderes
# dxt decoderes are based on Alex's code in below URL.
# base codes are modified for inplemented modules of blender 2.5
#
# http://code.google.com/r/andreasschiefer-pyglet-py3/source/browse/pyglet/image/codecs/s3tc.py?spec=svn018058c46c75c39280e8b29b39e0bdfb6db68720&r=018058c46c75c39280e8b29b39e0bdfb6db68720
#
# ----------------------------------------------------------------------------
# pyglet
# Copyright (c) 2006-2007 Alex Holkner
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions 
# are met:
#
#  * Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
#  * Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
#  * Neither the name of the pyglet nor the names of its
#    contributors may be used to endorse or promote products
#    derived from this software without specific prior written
#    permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ----------------------------------------------------------------------------

def decode_dxt1(data, width, height):
	# Decode to 24-bit RGB UNSIGNED_SHORT_8_8_8, and 8-bit Alpha UNSIGNED_SHORT_8
	# out[0]: RGB sequence [ num_pixs * 3 ]
	# out[1]: AAA sequence [ num_pix  * 3 ]
	out = [[0] * width * height * 3, [0] * width * height * 3]
	
	# Read 8 bytes at a time
	image_offset = 0
	
	for ib, block in enumerate( data ):
		( c0_lo, c0_hi, c1_lo, c1_hi, b0, b1, b2, b3 ) = struct.unpack( "<8B", block )
		color0 = c0_lo | c0_hi << 8
		color1 = c1_lo | c1_hi << 8
		bits   = b0 | b1 << 8 | b2 << 16 | b3 << 24
		
		r0 = ( ( color0 & 0x1f   )       )
		g0 = ( ( color0 & 0x7e0  ) >>  5 )
		b0 = ( ( color0 & 0xf800 ) >> 11 )
		r1 = ( ( color1 & 0x1f   )       )
		g1 = ( ( color1 & 0x7e0  ) >>  5 )
		b1 = ( ( color1 & 0xf800 ) >> 11 )
		
		# i is the dest ptr for this block
		i = image_offset
		for y in range(4):
			for x in range(4):
				code = bits & 0x3
				a = 255
				
				if code == 0:
					[r, g, b] = [r0, g0, b0]
				elif code == 1:
					[r, g, b] = [r1, g1, b1]
				elif code == 3 and color0 <= color1:
					[r, g, b] = [0]*3
					a = 0
				else:
					if code == 2 and color0 > color1:
						r = ( ( ( r0 * 2 ) + r1 ) // 3 )
						g = ( ( ( g0 * 2 ) + g1 ) // 3 )
						b = ( ( ( b0 * 2 ) + b1 ) // 3 )
					elif code == 3 and color0 > color1:
						r = ( ( r0 + ( r1 * 2 ) ) // 3 )
						g = ( ( g0 + ( g1 * 2 ) ) // 3 )
						b = ( ( b0 + ( b1 * 2 ) ) // 3 )
					else:
						assert code == 2 and color0 <= color1
						r = ( (r0 + r1) // 2 )
						g = ( (g0 + g1) // 2 )
						b = ( (b0 + b1) // 2 )
				
				out[0][i*3:i*3+3] = [ ( r << 3 ) + ( r >> 2 ), ( g << 2 ) + ( g >> 4 ), ( b << 3 ) + ( b >> 2 ) ]
				out[1][i*3:i*3+3] = [ a ]*3
				
				bits >>= 2
				i += 1
			i += width - 4
		
		# Move dest ptr to next 4x4 block
		advance_row = (image_offset + 4) % width == 0
		image_offset += width * 3 * advance_row + 4
	return out

def decode_dxt3(data, width, height):
	# Decode to 24-bit RGB UNSIGNED_SHORT_8_8_8, and 8-bit Alpha UNSIGNED_SHORT_8
	# out[0]: RGB sequence [ num_pixs * 3 ]
	# out[1]: AAA sequence [ num_pix  * 3 ]
	out = [[0] * width * height * 3, [0] * width * height * 3]
	
	# Read 16 bytes at a time
	image_offset = 0
	
	for ib, block in enumerate( data ):
		( a0, a1, a2, a3, a4, a5, a6, a7,
			c0_lo, c0_hi, c1_lo, c1_hi,
			b0, b1, b2, b3 ) = struct.unpack( "<16B", block )
		alpha = a0 | a1<<8 | a2<<16 | a3<<24 | a4<<32 | a5<<40 | a6<<48 | a7<<56
		color0 = c0_lo | c0_hi<<8
		color1 = c1_lo | c1_hi<<8
		bits   = b0 | b1<<8 | b2<<16 | b3<<24
		
		r0 = ( ( color0 & 0x1f   )       )
		g0 = ( ( color0 & 0x7e0  ) >>  5 )
		b0 = ( ( color0 & 0xf800 ) >> 11 )
		r1 = ( ( color1 & 0x1f   )       )
		g1 = ( ( color1 & 0x7e0  ) >>  5 )
		b1 = ( ( color1 & 0xf800 ) >> 11 )
		
		# i is the dest ptr for this block
		i = image_offset
		for y in range(4):
			for x in range(4):
				code = bits & 0x3
				a = alpha & 0xf
				
				if code == 0:
					[r, g, b] = [r0, g0, b0]
				elif code == 1:
					[r, g, b] = [r1, g1, b1]
				elif code == 3 and color0 <= color1:
					[r, g, b] = [0]*3
				else:
					if code == 2 and color0 > color1:
						r = ( ( ( r0 * 2 ) + r1 ) // 3 )
						g = ( ( ( g0 * 2 ) + g1 ) // 3 )
						b = ( ( ( b0 * 2 ) + b1 ) // 3 )
					elif code == 3 and color0 > color1:
						r = ( ( r0 + ( r1 * 2 ) ) // 3 )
						g = ( ( g0 + ( g1 * 2 ) ) // 3 )
						b = ( ( b0 + ( b1 * 2 ) ) // 3 )
					else:
						assert code == 2 and color0 <= color1
						r = ( (r0 + r1) // 2 )
						g = ( (g0 + g1) // 2 )
						b = ( (b0 + b1) // 2 )
				
				out[0][i*3:i*3+3] = [
					( r << 3 ) + r >> 2,
					( g << 2 ) + g >> 4,
					( b << 3 ) + b >> 2 ]
				
				out[0][i*3:i*3+3] = [ ( r << 3 ) + ( r >> 2 ), ( g << 2 ) + ( g >> 4 ), ( b << 3 ) + ( b >> 2 ) ]
				out[1][i*3:i*3+3] = [ ( a << 4 ) + a ]*3
				
				bits >>= 2
				alpha >>= 4
				i += 1
			i += width - 4
		
		# Move dest ptr to next 4x4 block
		advance_row = (image_offset + 4) % width == 0
		image_offset += width * 3 * advance_row + 4
	return out

#### END OF dxt decoderes

def WriteBMP32( filepath, image_data, alpha_data, width, height ):
	path = os.path.split( filepath )[0]
	if False == os.path.isdir( path ): os.makedirs( path )
	try:
		file = open( filepath, "wb" )
	except:
		pass#print( "-----\nError : failed to open " + filepath )
		return
	
	## BITMAPFILEHEADER  14 bytes
	# unsigned short bfType; 'BM'
	# unsigned long  bfSize;
	# unsigned short bfReserved1; always 0
	# unsigned short bfReserved2; always 0
	# unsigned long  bfOffBits;
	
	##BITMAPINFOHEADER  40 bytes
	# unsigned long  biSize;
	# long           biWidth;
	# long           biHeight;
	# unsigned short biPlanes; always 1
	# unsigned short biBitCount;
	# unsigned long  biCompression;
	# unsigned long  biSizeImage;
	# long           biXPixPerMeter;
	# long           biYPixPerMeter;
	# unsigned long  biClrUsed;
	# unsigned long  biClrImporant;
	
	num_pixs = width * height
	
	bfType      = 0x4d42  # 'BM'
	biSize      = 40      # 40 byte BITMAPINFOHEADER Windows Bitmap
	bfOffBits   = 14 + biSize
	biSizeImage = num_pixs * 4
	bfSize      = bfOffBits + biSizeImage
	biBitCount  = 32      # 24 bit/pix
	
	bfHeader = [ bfType, bfSize, 0, 0, bfOffBits ]
	biHeader = [ biSize, width, height, 1, biBitCount, 0, biSizeImage, 0, 0, 0, 0 ]
	file.write( struct.pack('<HI2HI', *bfHeader) )
	file.write( struct.pack('<I2i2H2I2i2I', *biHeader) )
	for il in range( height ):
		ofs = ( height - il - 1 ) * width * 3
		for ipix in range( width ):
			pix = ofs + ipix * 3
			file.write( struct.pack('<4B', image_data[pix], image_data[pix+1], image_data[pix+2], alpha_data[pix] ) )
	file.close()

def WriteBMP24( filepath, image_data, width, height ):
	#return
	path = os.path.split( filepath )[0]
	if False == os.path.isdir( path ): os.makedirs( path )
	try:
		file = open( filepath, "wb" )
	except:
		pass#print( "-----\nError : failed to open " + filepath )
		return
	
	## BITMAPFILEHEADER  14 bytes
	# unsigned short bfType; 'BM'
	# unsigned long  bfSize;
	# unsigned short bfReserved1; always 0
	# unsigned short bfReserved2; always 0
	# unsigned long  bfOffBits;
	
	##BITMAPINFOHEADER  40 bytes
	# unsigned long  biSize;
	# long           biWidth;
	# long           biHeight;
	# unsigned short biPlanes; always 1
	# unsigned short biBitCount;
	# unsigned long  biCompression;
	# unsigned long  biSizeImage;
	# long           biXPixPerMeter;
	# long           biYPixPerMeter;
	# unsigned long  biClrUsed;
	# unsigned long  biClrImporant;
	
	num_pixs = width * height
	
	bfType      = 0x4d42  # 'BM'
	biSize      = 40      # 40 byte BITMAPINFOHEADER Windows Bitmap
	bfOffBits   = 14 + biSize
	biSizeImage = num_pixs * 3
	bfSize      = bfOffBits + biSizeImage
	biBitCount  = 24      # 24 bit/pix
	
	bfHeader = [ bfType, bfSize, 0, 0, bfOffBits ]
	biHeader = [ biSize, width, height, 1, biBitCount, 0, biSizeImage, 0, 0, 0, 0 ]
	file.write( struct.pack('<HI2HI', *bfHeader) )
	file.write( struct.pack('<I2i2H2I2i2I', *biHeader) )
	for il in range( height ):
		ofs = ( height - il - 1 ) * width * 3
		file.write( struct.pack('<%dB' %( width * 3 ) , *image_data[ofs:ofs+width*3] ) )
	file.close()

def ExtractTXD( file, tex_list, tex_fld, extract = True, map = False, txd_path = None, alpha_mode = "BMP24" ):
	from . import import_dff
	#pass#print( "Read TXD" )
	texs = []
	
	import_dff.CheckHeader( file, 0x16 )  # Texture Dictionary
	import_dff.CheckHeader( file, 0x01 )  # Struct
	data = struct.unpack( "<2H", file.read( 4 ) )
	num_texs = data[0]
	#pass#print( "num_texs:", num_texs )
	
	for itex in range( num_texs ):
		## Texture Native: 92bytes + data
		#  4b - DWORD    - Platform ID
		#  2b - WORD     - Filter Flags
		#  1b - BYTE     - Texture Wrap V
		#  1b - BYTE     - Texture Wrap U
		# 32b - CHAR[32] - Diffuse Texture Name (null-terminated string)
		# 32b - CHAR[32] - Alpha Texture Name (null-terminated string)
		#  4b - DWORD    - Raster Format
		#  4b - DWORD    - Alpha/FOURCC
		#  2b - WORD     - Image Width
		#  2b - WORD     - Image Height
		#  1b - BYTE     - Bits Per Pixel (BPP)
		#  1b - BYTE     - Mipmap Count
		#  1b - BYTE     - Raster Type, always 0x04 ("Texture")
		#  1b - BYTE     - DXT Compression Type/Alpha
		#  4b - DWORD    - Image Data Size
		
		h = import_dff.CheckHeader( file, 0x15 )  # Texture Native
		endof_texnative = h.ofs + h.size
		import_dff.CheckHeader( file, 0x01 )  # Struct 
		data = struct.unpack( "<iH2B", file.read( 8 ) )
		pid     = data[0]
		filter  = data[1]
		wrap_v  = data[2]
		wrap_u  = data[3]
		tex_name= bytes.decode( file.read( 32 ), errors='replace' ).split( '\0' )[0].lower()
		alp_name= bytes.decode( file.read( 32 ), errors='replace' ).split( '\0' )[0].lower()
		data    = struct.unpack( "<2i2H4Bi", file.read( 20 ) )
		rst_fmt = data[0]
		fourcc  = data[1]
		width   = data[2]
		height  = data[3]
		is_alp  = data[7]
		size    = data[8]
		if 9 != pid:
			pass#print( "not GTASA format." )
			return
		
		if tex_name in tex_list:
			tex = ClassTXDTexture()
			tex.col_name = tex_name
			tex.col_path = txd_path + "\\" + tex.col_name
			if 1 == is_alp or 9 == is_alp:
				tex.alp_name = tex_name + "a"
				tex.alp_path = txd_path + "\\" + tex.alp_name
			texs.append( tex )
			
			if extract:
				tex.extracted = tex.col_path in map.tex_list
				if None != map and tex.extracted:
					map.tex_list[tex.col_path] = None
				if 1 == is_alp or 9 == is_alp:
					if tex.extracted: tex.extracted = tex.alp_path in map.tex_list
					if None != map and tex.extracted:
						map.tex_list[tex.alp_path] = None
				
				if tex.extracted:
					if map:
						map.counter += 1
						print( "Extracted ( " + str( map.counter ) + " / " + str( map.count_total ) + " ) : " + tex_name )
					else:
						print( "Extracted : " + tex_name )
				else:
					if map:
						map.counter += 1
						print( "Extract Texture( " + str( map.counter ) + " / " + str( map.count_total ) + " ) : " + tex_name )
					else:
						print( "Extract Texture : " + tex_name )
					image_data = [[],[]]
					
					## read raster data
					if 0x15 == fourcc or 0x16 == fourcc:
						num_pixs = width * height
						data = struct.unpack( '<%dB' % ( num_pixs * 4 ), file.read( num_pixs * 4 ) )
						for ip in range( num_pixs ):
							image_data[0] += data[ip * 4 : ip * 4 +3]
							image_data[1] += [ data[ip * 4 + 3] ] * 3
					
					if 0x31545844 == fourcc:
						data = []
						for pix in range( width * height // 16 ):
							data.append( file.read( 8 ) )
						image_data = decode_dxt1( data, width, height )
					
					if 0x33545844 == fourcc:
						data = []
						for pix in range( width * height // 16 ):
							data.append( file.read( 16 ) )
						image_data = decode_dxt3( data, width, height )
					
					## read to bmp file
					if "BMP24" == alpha_mode:
						WriteBMP24( tex_fld + "\\" + tex.col_path + ".bmp", image_data[0], width, height )
						if 1 == is_alp or 9 == is_alp:
							WriteBMP24( tex_fld + "\\" + tex.alp_path + ".bmp", image_data[1], width, height )
					elif "BMP32" == alpha_mode:
						if 1 == is_alp or 9 == is_alp:
							WriteBMP32( tex_fld + "\\" + tex.col_path + ".bmp", image_data[0], image_data[1], width, height )
						else:
							WriteBMP24( tex_fld + "\\" + tex.col_path + ".bmp", image_data[0], width, height )
			
			tex_list.remove( tex_name )
			if 0 == len( tex_list ): break
		file.seek( endof_texnative )
	return texs

